import typing

from graphql.error import format_error as format_graphql_error
from starlette import status
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, PlainTextResponse, Response

from .utils import get_playground_html


async def get_http_response(request: Request, execute: typing.Callable) -> Response:
    if request.method == "GET":
        html = get_playground_html(str(request.url))
        return HTMLResponse(html)

    if request.method == "POST":
        content_type = request.headers.get("Content-Type", "")

        if "application/json" in content_type:
            data = await request.json()
        else:
            return PlainTextResponse(
                "Unsupported Media Type",
                status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE,
            )
    else:
        return PlainTextResponse(
            "Method Not Allowed", status_code=status.HTTP_405_METHOD_NOT_ALLOWED
        )

    try:
        query = data["query"]
        variables = data.get("variables")
        operation_name = data.get("operationName")
    except KeyError:
        return PlainTextResponse(
            "No GraphQL query found in the request",
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    context = {"request": request}

    result = await execute(
        query, variables=variables, context=context, operation_name=operation_name
    )

    response_data = {"data": result.data}

    if result.errors:
        response_data["errors"] = [format_graphql_error(err) for err in result.errors]

    status_code = status.HTTP_400_BAD_REQUEST if result.errors else status.HTTP_200_OK

    return JSONResponse(response_data, status_code=status_code)



import inspect
import typing

import dataclasses
from graphql import GraphQLField, GraphQLInputField

from .constants import IS_STRAWBERRY_FIELD, IS_STRAWBERRY_INPUT
from .exceptions import MissingArgumentsAnnotationsError, MissingReturnAnnotationError
from .type_converter import REGISTRY, get_graphql_type_for_annotation
from .utils.dict_to_type import dict_to_type
from .utils.inspect import get_func_args
from .utils.lazy_property import lazy_property
from .utils.str_converters import to_camel_case, to_snake_case
from .utils.typing import (
    get_list_annotation,
    get_optional_annotation,
    is_list,
    is_optional,
)


class LazyFieldWrapper:
    """A lazy wrapper for a strawberry field.
    This allows to use cyclic dependencies in a strawberry fields:

    >>> @strawberry.type
    >>> class TypeA:
    >>>     @strawberry.field
    >>>     def type_b(self, info) -> "TypeB":
    >>>         from .type_b import TypeB
    >>>         return TypeB()
    """

    def __init__(
        self,
        obj,
        *,
        is_input=False,
        is_subscription=False,
        resolver=None,
        name=None,
        description=None,
        permission_classes=None
    ):
        self._wrapped_obj = obj
        self.is_subscription = is_subscription
        self.is_input = is_input
        self.field_name = name
        self.field_resolver = resolver
        self.field_description = description
        self.permission_classes = permission_classes or []

        if callable(self._wrapped_obj):
            self._check_has_annotations(self._wrapped_obj)

    def _check_has_annotations(self, func):
        # using annotations without passing from typing.get_type_hints
        # as we don't the actually types for the annotations
        annotations = func.__annotations__
        name = func.__name__

        if "return" not in annotations:
            raise MissingReturnAnnotationError(name)

        function_arguments = set(get_func_args(func)) - {"root", "self", "info"}

        arguments_annotations = {
            key: value
            for key, value in annotations.items()
            if key not in ["root", "info", "return"]
        }

        annotated_function_arguments = set(arguments_annotations.keys())
        arguments_missing_annotations = (
            function_arguments - annotated_function_arguments
        )

        if len(arguments_missing_annotations) > 0:
            raise MissingArgumentsAnnotationsError(name, arguments_missing_annotations)

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return getattr(self, attr)

        return getattr(self._wrapped_obj, attr)

    def __call__(self, *args, **kwargs):
        return self._wrapped_obj(self, *args, **kwargs)

    def _get_permissions(self):
        """
        Gets all permissions defined in the permission classes
        >>> strawberry.field(permission_classes=[IsAuthenticated])
        """
        return (permission() for permission in self.permission_classes)

    def _check_permissions(self, info):
        """
        Checks if the permission should be accepted and
        raises an exception if not
        """
        for permission in self._get_permissions():
            if not permission.has_permission(info):
                message = getattr(permission, "message", None)
                raise PermissionError(message)

    @lazy_property
    def field(self):
        return _get_field(
            self._wrapped_obj,
            is_input=self.is_input,
            is_subscription=self.is_subscription,
            name=self.field_name,
            description=self.field_description,
            check_permission=self._check_permissions,
        )


class strawberry_field(dataclasses.Field):
    """A small wrapper for a field in strawberry.

    You shouldn't be using this directly as this is used internally
    when using `strawberry.field`.

    This allows to use the following two syntaxes when using the type
    decorator:

    >>> class X:
    >>>     field_abc: str = strawberry.field(description="ABC")

    >>> class X:
    >>>     @strawberry.field(description="ABC")
    >>>     def field_a(self, info) -> str:
    >>>         return "abc"

    When calling this class as strawberry_field it creates a field
    that stores metadata (such as field description). In addition
    to that it also acts as decorator when called as a function,
    allowing us to us both syntaxes.
    """

    def __init__(
        self,
        *,
        is_input=False,
        is_subscription=False,
        resolver=None,
        name=None,
        description=None,
        metadata=None,
        permission_classes=None
    ):
        self.field_name = name
        self.field_description = description
        self.field_resolver = resolver
        self.is_subscription = is_subscription
        self.is_input = is_input
        self.field_permission_classes = permission_classes

        super().__init__(
            # TODO:
            default=dataclasses.MISSING,
            default_factory=dataclasses.MISSING,
            init=False,
            repr=True,
            hash=None,
            # TODO: this needs to be False when init is False
            # we could turn it to True when and if we have a default
            # probably can't be True when passing a resolver
            compare=False,
            metadata=metadata,
        )

    def __call__(self, wrap):
        setattr(wrap, IS_STRAWBERRY_FIELD, True)

        self.field_description = self.field_description or wrap.__doc__

        return LazyFieldWrapper(
            wrap,
            is_input=self.is_input,
            is_subscription=self.is_subscription,
            resolver=self.field_resolver,
            name=self.field_name,
            description=self.field_description,
            permission_classes=self.field_permission_classes,
        )


def convert_args(args, annotations):
    """Converts a nested dictionary to a dictionary of strawberry input types."""

    converted_args = {}

    for key, value in args.items():
        key = to_snake_case(key)
        annotation = annotations[key]

        # we don't need to check about unions here since they are not
        # yet supported for arguments.
        # see https://github.com/graphql/graphql-spec/issues/488

        is_list_of_args = False

        if is_optional(annotation):
            annotation = get_optional_annotation(annotation)

        if is_list(annotation):
            annotation = get_list_annotation(annotation)
            is_list_of_args = True

        if getattr(annotation, IS_STRAWBERRY_INPUT, False):
            if is_list_of_args:
                converted_args[key] = [dict_to_type(x, annotation) for x in value]
            else:
                converted_args[key] = dict_to_type(value, annotation)
        else:
            converted_args[key] = value

    return converted_args


def _get_field(
    wrap,
    *,
    is_input=False,
    is_subscription=False,
    name=None,
    description=None,
    check_permission=None
):
    name = wrap.__name__

    annotations = typing.get_type_hints(wrap, None, REGISTRY)
    parameters = inspect.signature(wrap).parameters
    field_type = get_graphql_type_for_annotation(annotations["return"], name)

    arguments_annotations = {
        key: value
        for key, value in annotations.items()
        if key not in ["info", "return"]
    }

    arguments = {
        to_camel_case(name): get_graphql_type_for_annotation(
            annotation, name, parameters[name].default != inspect._empty
        )
        for name, annotation in arguments_annotations.items()
    }

    def resolver(source, info, **args):
        if check_permission:
            check_permission(info)
        args = convert_args(args, arguments_annotations)
        return wrap(source, info, **args)

    field_params = {}

    if not is_input:
        field_params["args"] = arguments

        if is_subscription:

            def _resolve(event, info):
                if check_permission:
                    check_permission(info)
                return event

            field_params.update({"subscribe": resolver, "resolve": _resolve})
        else:
            field_params.update({"resolve": resolver})

    field_params["description"] = description or wrap.__doc__

    FieldType = GraphQLInputField if is_input else GraphQLField

    return FieldType(field_type, **field_params)


def field(
    wrap=None,
    *,
    name=None,
    description=None,
    resolver=None,
    is_input=False,
    is_subscription=False,
    permission_classes=None
):
    """Annotates a method or property as a GraphQL field.

    This is normally used inside a type declaration:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = strawberry.field(description="ABC")

    >>>     @strawberry.field(description="ABC")
    >>>     def field_with_resolver(self, info) -> str:
    >>>         return "abc"

    it can be used both as decorator and as a normal function.
    """

    field = strawberry_field(
        name=name,
        description=description,
        resolver=resolver,
        is_input=is_input,
        is_subscription=is_subscription,
        permission_classes=permission_classes,
    )

    # when calling this with parens we are going to return a strawberry_field
    # instance, so it can be used as both decorator and function.

    if wrap is None:
        return field

    # otherwise we run the decorator directly,
    # when called as @strawberry.field, without parens.

    return field(wrap)



from enum import EnumMeta

from graphql import GraphQLEnumType, GraphQLEnumValue
from graphql.utilities.schema_printer import print_type

from .exceptions import NotAnEnum
from .type_converter import REGISTRY


def _process_enum(cls, name=None, description=None):
    if not isinstance(cls, EnumMeta):
        raise NotAnEnum()

    if not name:
        name = cls.__name__

    REGISTRY[name] = cls

    def repr_(self):
        return print_type(self.field)

    setattr(cls, "__repr__", repr_)

    description = description or cls.__doc__

    cls.field = GraphQLEnumType(
        name=name,
        values=[(item.name, GraphQLEnumValue(item.value)) for item in cls],
        description=description,
    )

    return cls


def enum(_cls=None, *, name=None, description=None):
    """Registers the enum in the GraphQL type system.

    If name is passed, the name of the GraphQL type will be
    the value passed of name instead of the Enum class name.
    """

    def wrap(cls):
        return _process_enum(cls, name, description)

    if not _cls:
        return wrap

    return wrap(_cls)



import pathlib


def get_playground_html(request_path: str) -> str:
    here = pathlib.Path(__file__).parents[1]
    path = here / "static/playground.html"

    with open(path) as f:
        template = f.read()

    return template.replace("{{REQUEST_PATH}}", request_path)



import typing

from graphql import ExecutionResult, GraphQLSchema, parse
from graphql.subscription import subscribe as graphql_subscribe


async def subscribe(
    schema: GraphQLSchema,
    query: str,
    root_value: typing.Any = None,
    context_value: typing.Any = None,
    variable_values: typing.Dict[str, typing.Any] = None,
    operation_name: str = None,
) -> typing.Union[typing.AsyncIterator[ExecutionResult], ExecutionResult]:
    document = parse(query)

    return await graphql_subscribe(
        schema=schema,
        document=document,
        root_value=root_value,
        context_value=context_value,
        variable_values=variable_values,
        operation_name=operation_name,
    )



import click
import sys

import os
from starlette.applications import Starlette
from starlette.middleware.cors import CORSMiddleware

import importlib

import uvicorn

import hupper

from strawberry.asgi import GraphQL


@click.group()
def run():
    pass


@run.command("server")
@click.argument("module", type=str)
@click.option("-h", "--host", default="0.0.0.0", type=str)
@click.option("-p", "--port", default=8000, type=int)
def server(module, host, port):
    sys.path.append(os.getcwd())

    reloader = hupper.start_reloader("strawberry.cli.run", verbose=False)

    schema_module = importlib.import_module(module)

    reloader.watch_files([schema_module.__file__])

    app = Starlette(debug=True)

    app.add_middleware(
        CORSMiddleware, allow_headers=["*"], allow_origins=["*"], allow_methods=["*"]
    )

    graphql_app = GraphQL(schema_module.schema, debug=True)

    app.add_route("/graphql", graphql_app)
    app.add_websocket_route("/graphql", graphql_app)

    print(f"Running strawberry on http://{host}:{port}/graphql 🍓")

    uvicorn.run(app, host=host, port=port, log_level="error")



import typing


def is_list(annotation):
    """Returns True if annotation is a typing.List"""

    annotation_origin = getattr(annotation, "__origin__", None)

    return annotation_origin == list


def is_union(annotation):
    """Returns True if annotation is a typing.Union"""

    annotation_origin = getattr(annotation, "__origin__", None)

    return annotation_origin == typing.Union


def is_optional(annotation):
    """Returns True if the annotation is typing.Optional[SomeType]"""

    # Optionals are represented as unions

    if not is_union(annotation):
        return False

    types = annotation.__args__

    # A Union to be optional needs to have at least one None type
    return any([x == None.__class__ for x in types])  # noqa:E721


def get_optional_annotation(annotation):
    types = annotation.__args__
    non_none_types = [x for x in types if x != None.__class__]  # noqa:E721

    return non_none_types[0]


def get_list_annotation(annotation):
    return annotation.__args__[0]



from collections.abc import AsyncGenerator

from graphql import (
    GraphQLBoolean,
    GraphQLFloat,
    GraphQLID,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLString,
    GraphQLUnionType,
)

from .exceptions import UnallowedReturnTypeForUnion, WrongReturnTypeForUnion
from .scalars import ID
from .utils.typing import is_union


REGISTRY = {
    str: GraphQLString,
    int: GraphQLInt,
    float: GraphQLFloat,
    bool: GraphQLBoolean,
    ID: GraphQLID,
}


# TODO: make so that we don't pass force optional
# we use that when trying to get the type for a
# option field (which can either be a scalar or an object type)
def get_graphql_type_for_annotation(
    annotation, field_name: str, force_optional: bool = False
):
    # TODO: this might lead to issues with types that have a field value
    is_field_optional = force_optional

    if hasattr(annotation, "field"):
        graphql_type = annotation.field
    else:
        annotation_name = getattr(annotation, "_name", None)

        if annotation_name == "List":
            list_of_type = get_graphql_type_for_annotation(
                annotation.__args__[0], field_name
            )

            return GraphQLList(list_of_type)

        annotation_origin = getattr(annotation, "__origin__", None)

        if annotation_origin == AsyncGenerator:
            # async generators are used in subscription, we only need the yield type
            # https://docs.python.org/3/library/typing.html#typing.AsyncGenerator
            return get_graphql_type_for_annotation(annotation.__args__[0], field_name)

        elif is_union(annotation):
            types = annotation.__args__
            non_none_types = [x for x in types if x != None.__class__]  # noqa:E721

            # optionals are represented as Union[type, None]
            if len(non_none_types) == 1:
                is_field_optional = True
                graphql_type = get_graphql_type_for_annotation(
                    non_none_types[0], field_name, force_optional=True
                )
            else:
                is_field_optional = None.__class__ in types

                def _resolve_type(self, value, _type):
                    if not hasattr(self, "field"):
                        raise WrongReturnTypeForUnion(value.field_name, str(type(self)))

                    if self.field not in _type.types:
                        raise UnallowedReturnTypeForUnion(
                            value.field_name, str(type(self)), _type.types
                        )

                    return self.field

                # TODO: union types don't work with scalar types
                # so we want to return a nice error
                # also we want to make sure we have been passed
                # strawberry types
                graphql_type = GraphQLUnionType(
                    field_name, [type.field for type in types]
                )
                graphql_type.resolve_type = _resolve_type
        else:
            graphql_type = REGISTRY.get(annotation)

    if not graphql_type:
        raise ValueError(f"Unable to get GraphQL type for {annotation}")

    if is_field_optional:
        return graphql_type

    return GraphQLNonNull(graphql_type)



from functools import partial

import dataclasses
from graphql import GraphQLInputObjectType, GraphQLInterfaceType, GraphQLObjectType
from graphql.utilities.schema_printer import print_type

from .constants import IS_STRAWBERRY_FIELD, IS_STRAWBERRY_INPUT, IS_STRAWBERRY_INTERFACE
from .field import field, strawberry_field
from .type_converter import REGISTRY
from .utils.str_converters import to_camel_case


def _get_resolver(cls, field_name):
    class_field = getattr(cls, field_name, None)

    if class_field and getattr(class_field, "resolver", None):
        return class_field.resolver

    def _resolver(root, info):
        field_resolver = getattr(root or cls(), field_name, None)

        if getattr(field_resolver, IS_STRAWBERRY_FIELD, False):
            return field_resolver(root, info)

        elif field_resolver.__class__ is strawberry_field:
            # TODO: support default values
            return None

        return field_resolver

    return _resolver


def _process_type(cls, *, is_input=False, is_interface=False, description=None):
    name = cls.__name__
    REGISTRY[name] = cls

    def repr_(self):
        return print_type(self.field)

    setattr(cls, "__repr__", repr_)

    def _get_fields(wrapped):
        class_fields = dataclasses.fields(wrapped)

        fields = {}

        for class_field in class_fields:
            field_name = getattr(class_field, "field_name", None) or to_camel_case(
                class_field.name
            )
            description = getattr(class_field, "field_description", None)

            resolver = getattr(class_field, "field_resolver", None) or _get_resolver(
                cls, class_field.name
            )
            resolver.__annotations__["return"] = class_field.type

            fields[field_name] = field(
                resolver, is_input=is_input, description=description
            ).field

        strawberry_fields = {}

        for base in [cls, *cls.__bases__]:
            strawberry_fields.update(
                {
                    key: value
                    for key, value in base.__dict__.items()
                    if getattr(value, IS_STRAWBERRY_FIELD, False)
                }
            )

        for key, value in strawberry_fields.items():
            name = getattr(value, "field_name", None) or to_camel_case(key)

            fields[name] = value.field

        return fields

    if is_input:
        setattr(cls, IS_STRAWBERRY_INPUT, True)
    elif is_interface:
        setattr(cls, IS_STRAWBERRY_INTERFACE, True)

    extra_kwargs = {"description": description or cls.__doc__}

    if is_input:
        TypeClass = GraphQLInputObjectType
    elif is_interface:
        TypeClass = GraphQLInterfaceType
    else:
        TypeClass = GraphQLObjectType

        extra_kwargs["interfaces"] = [
            klass.field
            for klass in cls.__bases__
            if hasattr(klass, IS_STRAWBERRY_INTERFACE)
        ]

    wrapped = dataclasses.dataclass(cls, repr=False)
    wrapped.field = TypeClass(name, lambda: _get_fields(wrapped), **extra_kwargs)

    return wrapped


def type(cls=None, *, is_input=False, is_interface=False, description=None):
    """Annotates a class as a GraphQL type.

    Example usage:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = "ABC"
    """

    def wrap(cls):
        return _process_type(
            cls, is_input=is_input, is_interface=is_interface, description=description
        )

    if cls is None:
        return wrap

    return wrap(cls)


input = partial(type, is_input=True)
interface = partial(type, is_interface=True)



def lazy_property(fn):
    """Decorator that makes a property lazy-evaluated."""
    attr_name = "_lazy_" + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property



from dataclasses import is_dataclass


def dict_to_type(dict, cls):
    fields = cls.__dataclass_fields__

    kwargs = {}

    for name, field in fields.items():
        if is_dataclass(field.type):
            kwargs[name] = dict_to_type(dict.get(name, {}), field.type)
        else:
            kwargs[name] = dict.get(name)

    return cls(**kwargs)



from graphql import GraphQLSchema


# TODO: typings


class Schema(GraphQLSchema):
    def __init__(self, query, mutation=None, subscription=None):
        super().__init__(
            query=query.field,
            mutation=mutation.field if mutation else None,
            subscription=subscription.field if subscription else None,
        )



__version__ = "0.1.0"


from .enum import enum  # noqa
from .field import field  # noqa
from .mutation import mutation, subscription  # noqa
from .scalars import ID  # noqa
from .schema import Schema  # noqa
from .type import input, type, interface  # noqa
from .permission import BasePermission  # noqa
